/*
 SPDX-License-Identifier: GPL-3.0-or-later
 myMPD (c) 2018-2026 Juergen Mang <mail@jcgames.de>
 https://github.com/jcorporation/mympd
*/

/*! \file
 * \brief Script thread implementation
 */

#ifndef MYMPD_SCRIPTS_H
#define MYMPD_SCRIPTS_H

void *scripts_loop(void *arg_config);

#endif
