/*
 SPDX-License-Identifier: GPL-3.0-or-later
 myMPD (c) 2018-2026 Juergen Mang <mail@jcgames.de>
 https://github.com/jcorporation/mympd
*/

#define INCBIN_PREFIX
#define INCBIN_STYLE INCBIN_STYLE_SNAKE

#include "compile_time.h"
#include "dist/incbin/incbin.h"

//compressed assets
INCBIN(sw_js, "/build/mympd/src/build/htdocs/sw.js.gz");
INCBIN(mympd_webmanifest, "/build/mympd/src/build/htdocs/mympd.webmanifest.gz");
INCBIN(index_html, "/build/mympd/src/build/htdocs/index.html.gz");
INCBIN(coverimage_transparent_svg, "/build/mympd/src/build/htdocs/assets/coverimage-transparent.svg.gz");
INCBIN(coverimage_notavailable_svg, "/build/mympd/src/build/htdocs/assets/coverimage-notavailable.svg.gz");
INCBIN(coverimage_stream_svg, "/build/mympd/src/build/htdocs/assets/coverimage-stream.svg.gz");
INCBIN(coverimage_booklet_svg, "/build/mympd/src/build/htdocs/assets/coverimage-booklet.svg.gz");
INCBIN(coverimage_folder_svg, "/build/mympd/src/build/htdocs/assets/coverimage-folder.svg.gz");
INCBIN(coverimage_mympd_svg, "/build/mympd/src/build/htdocs/assets/coverimage-mympd.svg.gz");
INCBIN(coverimage_playlist_svg, "/build/mympd/src/build/htdocs/assets/coverimage-playlist.svg.gz");
INCBIN(coverimage_smartpls_svg, "/build/mympd/src/build/htdocs/assets/coverimage-smartpls.svg.gz");
INCBIN(mympd_background_dark_svg, "/build/mympd/src/build/htdocs/assets/mympd-background-dark.svg.gz");
INCBIN(mympd_background_light_svg, "/build/mympd/src/build/htdocs/assets/mympd-background-light.svg.gz");
INCBIN(combined_css, "/build/mympd/src/build/htdocs/css/combined.css.gz");
INCBIN(combined_js, "/build/mympd/src/build/htdocs/js/combined.js.gz");
INCBIN(MaterialIcons_Regular_woff2, "/build/mympd/src/build/htdocs/assets/MaterialIcons-Regular.woff2");
INCBIN(ligatures_json, "/build/mympd/src/build/htdocs/assets/ligatures.json.gz");
//translation files
#ifdef I18N_bg_BG
    INCBIN(i18n_bg_BG_json, "/build/mympd/src/build/htdocs/assets/i18n/bg-BG.json.gz");
#endif
#ifdef I18N_de_DE
    INCBIN(i18n_de_DE_json, "/build/mympd/src/build/htdocs/assets/i18n/de-DE.json.gz");
#endif
#ifdef I18N_en_US
    INCBIN(i18n_en_US_json, "/build/mympd/src/build/htdocs/assets/i18n/en-US.json.gz");
#endif
#ifdef I18N_es_AR
    INCBIN(i18n_es_AR_json, "/build/mympd/src/build/htdocs/assets/i18n/es-AR.json.gz");
#endif
#ifdef I18N_es_ES
    INCBIN(i18n_es_ES_json, "/build/mympd/src/build/htdocs/assets/i18n/es-ES.json.gz");
#endif
#ifdef I18N_es_VE
    INCBIN(i18n_es_VE_json, "/build/mympd/src/build/htdocs/assets/i18n/es-VE.json.gz");
#endif
#ifdef I18N_fi_FI
    INCBIN(i18n_fi_FI_json, "/build/mympd/src/build/htdocs/assets/i18n/fi-FI.json.gz");
#endif
#ifdef I18N_fr_FR
    INCBIN(i18n_fr_FR_json, "/build/mympd/src/build/htdocs/assets/i18n/fr-FR.json.gz");
#endif
#ifdef I18N_it_IT
    INCBIN(i18n_it_IT_json, "/build/mympd/src/build/htdocs/assets/i18n/it-IT.json.gz");
#endif
#ifdef I18N_ja_JP
    INCBIN(i18n_ja_JP_json, "/build/mympd/src/build/htdocs/assets/i18n/ja-JP.json.gz");
#endif
#ifdef I18N_ko_KR
    INCBIN(i18n_ko_KR_json, "/build/mympd/src/build/htdocs/assets/i18n/ko-KR.json.gz");
#endif
#ifdef I18N_nl_NL
    INCBIN(i18n_nl_NL_json, "/build/mympd/src/build/htdocs/assets/i18n/nl-NL.json.gz");
#endif
#ifdef I18N_pl_PL
    INCBIN(i18n_pl_PL_json, "/build/mympd/src/build/htdocs/assets/i18n/pl-PL.json.gz");
#endif
#ifdef I18N_ru_RU
    INCBIN(i18n_ru_RU_json, "/build/mympd/src/build/htdocs/assets/i18n/ru-RU.json.gz");
#endif
#ifdef I18N_zh_Hans
    INCBIN(i18n_zh_Hans_json, "/build/mympd/src/build/htdocs/assets/i18n/zh-Hans.json.gz");
#endif
//uncompressed assets
INCBIN(appicon_192_png, "/build/mympd/src/build/htdocs/assets/appicon-192.png");
INCBIN(appicon_512_png, "/build/mympd/src/build/htdocs/assets/appicon-512.png");
