/*
 SPDX-License-Identifier: GPL-3.0-or-later
 myMPD (c) 2018-2026 Juergen Mang <mail@jcgames.de>
 https://github.com/jcorporation/mympd
*/

/*! \file
 * \brief myMPD database API
 */

#ifndef MYMPD_API_DATABASE_H
#define MYMPD_API_DATABASE_H

#include "dist/sds/sds.h"
#include "src/lib/api.h"
#include "src/lib/config/mympd_state.h"

#include <stdbool.h>

sds mympd_api_database_update(struct t_partition_state *partition_state, sds buffer, enum mympd_cmd_ids cmd_id, unsigned request_id, sds path);

#endif
